/*
 * Copyright (c) 2014 Citrix.
 * Written by Frediano Ziglio <frediano.ziglio@citrix.com>.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if EXACT_TESTING
#  undef NDEBUG
#  include "mock_defs.h"
#  include <assert.h>
#  define dbg(s) s
#else
#  include "defs.h"
#  define dbg(s) do {} while(0)
#endif

#include <stdarg.h>
#include <stddef.h>

#include "structs.h"

static void
assure_space(print_status *status, size_t needed)
{
	/* assert(needed <= sizeof(status->buffer)); */
	if (needed > status->dest_end - status->dest) {
		tprints(status->buffer);
		status->buffer[0] = 0;
		status->dest = status->buffer;
	}
}

void
status_flush(print_status *status)
{
	assure_space(status, sizeof(status->buffer));
}

void
status_printf(print_status *status, const char *fmt, ...)
{
	int l;
	va_list ap;

	assure_space(status, 64);
	va_start(ap, fmt);
	l = vsprintf(status->dest, fmt, ap);
	va_end(ap);
	status->dest += l;
}

static void
print_void(print_status *status, const void *data, const void *info)
{
}

const type_base type_void = {
	{ print_void, 0 }
};

void
print_struct(print_status *status, const void *_data, const void *_info)
{
	const unsigned char *data = (const unsigned char *)_data;
	const type_struct *info = (const type_struct *)_info;

	char sep = '{';
	const struct_item *item = info->items;
	// foreach field
	for (; item->name; ++item, sep = ',') {
		// print field, value
		status_printf(status, "%c%s=", sep, item->name);
		item->info->print(status, (const void*) (data+item->off), item->info);
	}
	status_printf(status, "}");
}

void
print_fixed_array(print_status *status, const void *_data, const void *_info)
{
	const unsigned char *data = (const unsigned char *)_data;
	const type_fixed_array *info = (const type_fixed_array *)_info;

	const unsigned char *const end = data + info->base.size;
	const base_type *item = info->item_type;
	const char *sep = NULL;

	status_printf(status, "[");
	for (;data < end; data += item->size, sep = ",") {
		if (sep) status_printf(status, sep);
		item->print(status, data, item);
	}
	status_printf(status, "]");
}

void
print_ptr_array(print_status *status, const void *_data, const void *_info)
{
	const unsigned char *data_ptr = (const unsigned char *)_data;
	const type_ptr_array *info = (const type_ptr_array *)_info;

	const base_type *item = info->item_type;
	int num;
	char *end, *data, *buf;
	long arg = *((long*)data_ptr);
	const char *sep = NULL;

	// TODO find another way to get the number
	num = *((int*)(data_ptr+info->num_off));
	if (num < 0) num = 0;
	if (num > 1024) num = 1024;

	data = buf = (char*) malloc(item->size * num);
	if (!buf || umoven(status->tcp, arg, item->size * num, buf) < 0) {
		status_printf(status, "BADPTR:%#lx", arg);
		free(buf);
		return;
	}
	end  = data + item->size * num;
	status_printf(status, "[");
	for (;data < end; data += item->size, sep = ",") {
		if (sep) status_printf(status, sep);
		item->print(status, data, item);
	}
	status_printf(status, "]");
	free(buf);
}

#if 0
typedef struct {
	unsigned value;
	const char *name;
	const base_type *type;
	char output;	/* != 0 if output */
} switch_item;

typedef struct {
	DATA_START;
	const switch_item *items;
	ssize_t data_off;	/* offset to data from number */
} type_switch;

static void
print_switch(print_status *status, const void *_data, const void *_info)
{
	const unsigned char *data = (const unsigned char *)_data;
	const type_switch *info = (const type_switch *)_info;

	/* check if we have */
	unsigned val = *((unsigned *) data);
	const switch_item *item = info->items;
	/* TODO sort items */
	for (; item->name; ++item) {
		if (item->value != val)
			continue;

		/* check if proper state */
//		if (status->entering == item->output)
//			return;

		/* found !*/
		status_printf(status, "%s,", item->name);
		item->type->print(status, data+info->data_off, item->type);
		return;
	}

	/* not found */
	if (status->entering)
		status_printf(status, "UNKNOWN(%lu)", val);
}
#endif

static void
print_guid(print_status *status, const void *_data, const void *dummy)
{
	struct guid {
		uint32_t n1;
		uint16_t n2, n3;
		uint8_t  n4[8];
	} *data = (struct guid*) _data;
	status_printf(status, "%08x-%04x-%04x-%02x%02x%02x%02x%02x%02x%02x%02x", data->n1, data->n2, data->n3,
		data->n4[0], data->n4[1], data->n4[2], data->n4[3], data->n4[4], data->n4[5], data->n4[6], data->n4[7]);
}

const type_base type_guid = {
	{ print_guid, 16 }
};

typedef unsigned long seq_type;
#define SEQ_FMT "%lu"
typedef long diff_type;
#define DIFF_FMT "%ld"

static void
print_ulong_seq(print_status *status, const seq_type *data, int seq, diff_type offset, const char **sep)
{
	if (seq < 3) {
		data -= seq;
		for (; seq; --seq, ++data) {
			status_printf(status, "%s", *sep);
       			status_printf(status, SEQ_FMT, *data);
			*sep = ",";
		}
		return;
	}

	status_printf(status, "%s", *sep);
	*sep = ",";
	if (offset == 0) {
		status_printf(status, "%d times " SEQ_FMT, seq, data[-seq]);
	} else if (offset == 1) {
		status_printf(status, SEQ_FMT " to " SEQ_FMT, data[-seq], data[-1]);
	} else {
		status_printf(status, SEQ_FMT " to " SEQ_FMT " step " DIFF_FMT, data[-seq], data[-1], offset);
	}
}

void
print_ulong_array(print_status *status, const void *_data, const void *_info)
{
	const unsigned char *data_ptr = (const unsigned char *)_data;
	const type_ulong_array *info = (const type_ulong_array *)_info;

	int num, seq;
	unsigned long *end, *data, *buf;
	unsigned long offset;
	long arg = *((long*)data_ptr);
	const char *sep = "";
	dbg(int count = 0);

	// TODO find another way to get the number
	num = *((int*)(data_ptr+info->num_off));
	if (num < 0) num = 0;
	if (num > 1024) num = 1024;

	data = buf = (unsigned long*) malloc(sizeof(long) * num);
	if (!buf || umoven(status->tcp, arg, sizeof(long) * num, (char *) buf) < 0) {
		status_printf(status, "BADPTR:%#lx", arg);
		free(buf);
		return;
	}

	dbg(printf("-------\n"));
	end  = data + num;
	status_printf(status, "[");
	seq = 0;
	offset = 0;
	for (;data < end; ++data) {
		dbg(assert(data < end));
		if (seq == 0) {
			++seq;
		} else if (seq == 1) {
			dbg(assert(data - 1 >= buf));
			offset = data[0] - data[-1];
			++seq;
		} else if (data[0] == data[-1] + offset) {
			++seq;
		} else {
			if (seq == 2) { --data, --seq; }
			dbg(fprintf(stderr, "seq %d offset " DIFF_FMT " data pos %ld data[0] " SEQ_FMT "\n", seq, offset, data - buf, data[0]));
			/* print elements in the sequence and start again */
			dbg(count += seq);
			dbg(assert(data - seq >= buf));
			dbg(assert(data < end));
			print_ulong_seq(status, data, seq, offset, &sep);
			seq = 0;
			dbg(assert(data > buf));
			--data;
		}
	}
	dbg(count += seq);
	dbg(assert(data - seq >= buf));
	dbg(assert(data == end));
	print_ulong_seq(status, data, seq, offset, &sep);
	dbg(fprintf(stderr, "seq %d\n", seq));
	dbg(assert(num == count));
	status_printf(status, "]");
	free(buf);
}

